#ifndef __GBA_HPP__
#define __GBA_HPP__

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned long u32;
typedef signed char s8;
typedef short s16;
typedef long s32;

#define MEM_IN_EWRAM __attribute__ ((section (".ewram")))
#define MEM_IN_IWRAM __attribute__ ((section (".iwram")))
#define MEM_FUNC_IN_IWRAM __attribute__ ((section (".iwram"), long_call))

#define RGB(r,g,b) (u16)(r + (g << 5) + (b << 10))

#define VID_BUFFERA     0x06000000
#define VID_BUFFERB     0x0600A000

// LCD I/O Registers
#define REG_DISPCNT     *(u32*)0x4000000
#define REG_DISPSTAT    *(volatile u16*)0x4000004
#define REG_VCOUNT      *(u16*)0x4000006
#define REG_BG0CNT      *(volatile u16*)0x4000008
#define REG_BG1CNT      *(volatile u16*)0x400000A
#define REG_BG2CNT      *(volatile u16*)0x400000C
#define REG_BG3CNT      *(volatile u16*)0x400000E

#define REG_BG0HOFS     *(volatile u16*)0x4000010
#define REG_BG0VOFS     *(volatile u16*)0x4000012
#define REG_BG1HOFS     *(volatile u16*)0x4000014
#define REG_BG1VOFS     *(volatile u16*)0x4000016
#define REG_BG2HOFS     *(volatile u16*)0x4000018
#define REG_BG2VOFS     *(volatile u16*)0x400001A
#define REG_BG3HOFS     *(volatile u16*)0x400001C
#define REG_BG3VOFS     *(volatile u16*)0x400001E

#define REG_BG2X        *(volatile u32*)0x4000028
#define REG_BG2Y        *(volatile u32*)0x400002C
#define REG_BG2PA       *(volatile u16*)0x4000020
#define REG_BG2PB       *(volatile u16*)0x4000022
#define REG_BG2PC       *(volatile u16*)0x4000024
#define REG_BG2PD       *(volatile u16*)0x4000026

#define REG_BG3X        *(volatile u32*)0x4000038
#define REG_BG3Y        *(volatile u32*)0x400003C
#define REG_BG3PA       *(volatile u16*)0x4000030
#define REG_BG3PB       *(volatile u16*)0x4000032
#define REG_BG3PC       *(volatile u16*)0x4000034
#define REG_BG3PD       *(volatile u16*)0x4000036

#define REG_BLDCNT      *(volatile u16*)0x4000050
#define REG_BLDALPHA    *(volatile u16*)0x4000052
#define REG_BLDY        *(volatile u16*)0x4000054

#define BLD_1_BG0       0x0001
#define BLD_1_BG1       0x0002
#define BLD_1_BG2       0x0004
#define BLD_1_BG3       0x0008
#define BLD_1_OBJ       0x0010
#define BLD_1_BD        0x0020
#define BLD_NONE        0x0000
#define BLD_ALPHA       0x0040
#define BLD_BRIGHT      0x0080
#define BLD_DARK        0x00C0
#define BLD_2_BG0       0x0100
#define BLD_2_BG1       0x0200
#define BLD_2_BG2       0x0400
#define BLD_2_BG3       0x0800
#define BLD_2_OBJ       0x1000
#define BLD_2_BD        0x2000

#define VID_BACKBUFFER  0x0010
#define VID_HBLANK      0x0020
#define VID_1DMAP       0x0040
#define VID_BLANK       0x0080
#define VID_BG0         0x0100
#define VID_BG1         0x0200
#define VID_BG2         0x0400
#define VID_BG3         0x0800
#define VID_OBJ         0x1000
#define VID_WIN0        0x2000
#define VID_WIN1        0x4000
#define VID_OBJWIN      0x8000
#define SetMode(mode)   REG_DISPCNT = (mode)

// Interrupt Registers

#define REG_IME         *(u16*)0x4000208
#define REG_IE          *(u16*)0x4000200
#define REG_IF          *(u16*)0x4000202
#define REG_INTERRUPT   *(u32*)0x3007FFC

#define INT_VBLANK      0x0001
#define INT_HBLANK      0x0002
#define INT_VCOUNT      0x0004
#define INT_TIM0        0x0008
#define INT_TIM1        0x0010
#define INT_TIM2        0x0020
#define INT_TIM3        0x0040
#define INT_SIO         0x0080
#define INT_DMA0        0x0100
#define INT_DMA1        0x0200
#define INT_DMA2        0x0400
#define INT_DMA3        0x0800
#define INT_KEY         0x1000
#define INT_GAME        0x2000

// OAM Data

#define A0_RSENABLE     0x0100
#define A0_DOUBLE       0x0200
#define A0_DISABLE      0x0200
#define A0_NORMAL       0x0000
#define A0_SEMITRANS    0x0400
#define A0_OBJWIN       0x0800
#define A0_MOSAIC       0x1000
#define A0_256PAL       0x2000
#define A0_SQUARE       0x0000
#define A0_HORZ         0x4000
#define A0_VERT         0x8000

#define A1_RS(x)        (x << 9)
#define A1_HFLIP        0x1000
#define A1_VFLIP        0x2000
#define A1_SIZE(x)      (x << 14)

#define A2_PRIORITY(x)  (x << 10)
#define A2_PALETTE(x)   (x << 12)

typedef struct sOBJdata {
    u16 a0;
    u16 a1;
    u16 a2;
    u16 a3;
} OBJdata;

// Timer Registers
#define REG_TM0D        *(volatile u16*)0x4000100
#define REG_TM0CNT      *(volatile u16*)0x4000102
#define REG_TM1D        *(volatile u16*)0x4000104
#define REG_TM1CNT      *(volatile u16*)0x4000106
#define REG_TM2D        *(volatile u16*)0x4000108
#define REG_TM2CNT      *(volatile u16*)0x400010A
#define REG_TM3D        *(volatile u16*)0x400010C
#define REG_TM3CNT      *(volatile u16*)0x400010E

#define TIMER_SCALE_1       0x0000
#define TIMER_SCALE_64      0x0001
#define TIMER_SCALE_256     0x0002
#define TIMER_SCALE_1024    0x0003
#define TIMER_COUNT_UP      0x0004
#define TIMER_OVERFLOW      0x0040
#define TIMER_ENABLE        0x0080

// DMA Transfer Channels
#define REG_DMA0SAD     *(volatile u32*)0x40000B0
#define REG_DMA0DAD     *(volatile u32*)0x40000B4
#define REG_DMA0CNT_L   *(volatile u16*)0x40000B8
#define REG_DMA0CNT_H   *(volatile u16*)0x40000BA
#define REG_DMA1SAD     *(volatile u32*)0x40000BC
#define REG_DMA1DAD     *(volatile u32*)0x40000C0
#define REG_DMA1CNT_L   *(volatile u16*)0x40000C8
#define REG_DMA1CNT_H   *(volatile u16*)0x40000C6
#define REG_DMA2SAD     *(volatile u32*)0x40000C8
#define REG_DMA2DAD     *(volatile u32*)0x40000CC
#define REG_DMA2CNT_L   *(volatile u16*)0x40000D0
#define REG_DMA2CNT_H   *(volatile u16*)0x40000D2
#define REG_DMA3SAD     *(volatile u32*)0x40000D4
#define REG_DMA3DAD     *(volatile u32*)0x40000D8
#define REG_DMA3CNT_L   *(volatile u16*)0x40000DC
#define REG_DMA3CNT_H   *(volatile u16*)0x40000DE

#define DMA_DEST_FIXED  0x0040
#define DMA_REPEAT      0x0200
#define DMA_16          0x0000
#define DMA_32          0x0400
#define DMA_ENABLE      0x8000
#define DMA_SPECIAL     0x3000
#define DMA_TIMING_IMMEDIATE 0x0000


// Sound Registers

#define REG_SGCNT0_L    *(volatile u16*)0x4000080
#define REG_SGCNT0_H    *(volatile u16*)0x4000082
#define REG_SGCNT1      *(volatile u16*)0x4000084
#define REG_SGBIAS      *(volatile u16*)0x4000088
#define REG_SGFIFOA     *(volatile u32*)0x40000A0
#define REG_SGFIFOB     *(volatile u32*)0x40000A4

#define DSOUND_A_VOL            0x0004
#define DSOUND_B_VOL            0x0008
#define DSOUND_A_RIGHT_CHANNEL  0x0100
#define DSOUND_A_LEFT_CHANNEL   0x0200
#define DSOUND_A_TIM1           0x0400
#define DSOUND_A_FIFO_RESET     0x0800
#define DSOUND_B_RIGHT_CHANNEL  0x1000
#define DSOUND_B_LEFT_CHANNEL   0x2000
#define DSOUND_B_TIM1           0x4000
#define DSOUND_B_FIFO_RESET     0x8000

#define SOUND_MASTER_ENABLE     0x0080
#define SOUND_FIFO_A            0x40000A0
#define SOUND_FIFO_B            0x40000A4

// Button Values

#define BTN_A                   0x0001
#define BTN_B                   0x0002
#define BTN_SELECT              0x0004
#define BTN_START               0x0008
#define BTN_RIGHT               0x0010
#define BTN_LEFT                0x0020
#define BTN_UP                  0x0040
#define BTN_DOWN                0x0080
#define BTN_R                   0x0100
#define BTN_L                   0x0200

// SRAM registers

#define GAMEPAK_RAM  ((volatile u8*)0x0E000000)
#define WR_U32(a,b,c,d) (((u32)a << 24) | ((u32)b << 16) | ((u32)c << 8) | (u32)d)
#define RD_U32(i,x)     (u8)(i >> (x*8))

#endif
