/****************************************************************
 *
 * Functions put into IWRAM must be in a seperate file
 *
 ***************************************************************/

#include "gba.hpp"
#include "game.hpp"
#include "iwram.hpp"

extern u32 highscores[10];
extern u32 mixBufferP;
extern bool readA;
extern volatile MixerData Mixer;
extern u8 pvala[16];
extern u8 pvalb[16];
extern u8 *pval;
extern bool nextBar;
extern volatile u8 miss;
extern volatile u8 *pSaveMemory;

// Interrupt Handler
void IntHandler(void) {
    //REG_IME = 0x00;
    u16 Int_Flag = REG_IF;

    if(REG_IF & INT_TIM1) {
        REG_DMA1CNT_H = 0;
        REG_DMA1SAD = mixBufferP;
        REG_DMA1DAD = SOUND_FIFO_A;
        REG_DMA1CNT_H = DMA_DEST_FIXED | DMA_REPEAT | DMA_32 |
            DMA_SPECIAL | DMA_ENABLE;
        readA = true;
    } else if(REG_IF & INT_VBLANK) {
        miss++;
    } else if(REG_IF & INT_TIM2) {
        nextBar = true;
        Mixer.bar++;
        Mixer.bar %= 4;
        if(pval == pvala) pval = pvalb;
        else pval = pvala;
    }
    REG_IF = Int_Flag;
    //REG_IME = 0x01;
}

// Reading and writing to Game Pak SRAM
// Bus width is 8-bits, so larger data types need to be broken up

void ReadHighScores(void) {
    u8 a,b,c,d;
    for( u8 i = 0; i < 10; i++ ) {
        a = pSaveMemory[i*4];
        b = pSaveMemory[i*4+1];
        c = pSaveMemory[i*4+2];
        d = pSaveMemory[i*4+3];
        highscores[i] = WR_U32(a,b,c,d);
    }
}

void WriteHighScores(void) {
    for( u8 i = 0; i < 10; i++ ) {
        pSaveMemory[i*4] = RD_U32(highscores[i],3);
        pSaveMemory[i*4+1] = RD_U32(highscores[i],2);
        pSaveMemory[i*4+2] = RD_U32(highscores[i],1);
        pSaveMemory[i*4+3] = RD_U32(highscores[i],0);
    }
}

