/************************************************************************
 * Rez Advance
 * Copyright Mark Johnson, 2005
 ************************************************************************/

#include "gba.hpp"
#include "game.hpp"
#include "iwram.hpp"
#include "leveldata.hpp"

#define MIX_BUFFER_SIZE 4096
#define MIX_BUFFER_SIZE_2 2048

OBJdata objs[128];

// Objects in OAM
#define OBJ(x)          objs[x]
#define OBJ_NUM(x)      objs[x]
#define OBJ_SELF        objs[10]
#define OBJ_HEALTH(x)   objs[11+x]
#define OBJ_TARGET      objs[16]
#define OBJ_PLAYER      objs[17]
#define OBJ_ENEMY(x)    objs[32+x]
#define OBJ_WALL(x)     objs[120+x]
#define OBJ_MAX_ENEMY   96

// Sprite offsets
#define S_TARGET        512
#define S_TARGETING     516
#define S_NUM(x)        (520 + 2*(x))
#define S_HEALTH_ON0    552
#define S_HEALTH_OFF0   554
#define S_HEALTH_ON1    556
#define S_HEALTH_OFF1   558
#define S_HEALTH_ON2    556
#define S_HEALTH_OFF2   558
#define S_HEALTH_ON3    556
#define S_HEALTH_OFF3   558
#define S_HEALTH_ON4    560
#define S_HEALTH_OFF4   562
#define S_SELF          896
#define S_SELF_HIT      904
#define S_SELF_KILL(x)  (912 + 8*(x))
#define S_ENEMY         640
#define S_ENEMY_TARG    656
#define S_MISSILE_0     576
#define S_MISSILE_1     580
#define S_MISSILE_2     584
#define S_MISSILE_3     588
#define TARGET_MISSILE(X) (X + 16)

// Rotation/Scaling Matrix
#define RS_PA(i)        objs[i*4].a3
#define RS_PB(i)        objs[i*4 + 1].a3
#define RS_PC(i)        objs[i*4 + 2].a3
#define RS_PD(i)        objs[i*4 + 3].a3

#define MAX_DEPTH       200
#define MAX_ENEMIES     32
#define MAX_MISSILES    24

#define EN_NORMAL       0x00
#define EN_TARGET       0x01
#define EN_KILLWAIT     0x02
#define EN_KILL         0x04

#define MS_NORMAL       0x00
#define MS_TARGET       0x01
#define MS_HIT          0x02

#define KILL_ENEMY      0x00
#define KILL_MISSILE    0x01
#define KILL_QUEUE_MAX  16


// GBA Memory Addresses
u16 *videoBuffer        = (u16*)VID_BUFFERB;
u16 *FrontBuffer        = (u16*)VID_BUFFERA;
u16 *BackBuffer         = (u16*)VID_BUFFERB;
u16 *ObjTiles           = (u16*)0x06014000;
volatile u16 *vcount    = (volatile u16*)0x4000006;
u16 *palette            = (u16*)0x5000000;
u16 *pptr               = (u16*)0x5000002;
u16 *objPalette         = (u16*)0x5000200;
volatile u16 *beatTimer = (volatile u16*)0x4000108;
volatile u32 *BUTTONS   = (volatile u32 *)0x04000130;
u16 *OAM                = (u16*)0x7000000;
volatile u8 *pSaveMemory = GAMEPAK_RAM;

// Backgroun line data
#include "linecoord.h"
const u16 *linea[] = {linea1, linea2, linea3, linea4};
const u16 *lineb[] = {lineb1, lineb2, lineb3, lineb4};
#define linecol(j) (90 - j/2)

// Bitmaps for game menus
extern u8 gameover[];
extern u16 gameoverpal[];
extern u8 highscore[];
extern u16 highscorepal[];
extern u8 menuitems[];
extern u16 menupal[];
extern u8 title[];
extern u16 titlepal[];
extern u16 help1pal[];
extern u8 help1[];
extern u16 help2pal[];
extern u8 help2[];
extern u16 help3pal[];
extern u8 help3[];

const u8 numhelp = 3;

u8 *helpmenubmp[numhelp] = {help1,help2,help3};
u16 *helppal[numhelp] = {help1pal,help2pal,help3pal};

u16 pmem[256];

u32 highscores[10];

// Global variables for game data
// This is really ugly, should be converted into structs
// that are passed around between functions
u32 score = 0;
u8 pcnt = 20;
u8 health = 5;
u8 targets = 0;
u8 tokill = 0;
u8 killidx = 0;
u8 kidxmix = 0;
u8 killidxback = 0;
u8 tohit = 0;
u8 hitidx = 0;
u8 hidxmix = 0;
u8 hitidxback = 0;
u16 btns = 0xffff;
u16 hitmissiles = 0;
u16 hitenemies = 0;
u16 totenemies = 0;
u16 attacks = 0;

// Linked Lists made inside arrays
EnemyData enemies[MAX_ENEMIES];
MissileData missiles[MAX_MISSILES];

// Queues for things waiting to be destroy
killdata targetidxs[4];
killdata killidxs[KILL_QUEUE_MAX];
killdata hitidxs[KILL_QUEUE_MAX];

s8 enemyfptr = -1;
s8 enemybptr = -1;
u8 numEnemies = 0;
u8 numMaxEnemies = 10;

s8 missilefptr = -1;
s8 missilebptr = -1;
u8 numMissiles = 0;
u8 numMaxMissiles = 6;

// Coordinates for enemies
const s16 QUADX[12] = {0, 60, 120, 180, 0, 180, 0, 180, 0, 60, 120, 180};
const s16 QUADY[12] = {0, 0, 0, 0, 40, 40, 80, 80, 120, 120, 120, 120};

// Double buffer for palette shifts
u8 pvala[16] = {20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20};
u8 pvalb[16] = {20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20};
u8 *pval = pvala;
bool nextBar = false;

u32 samLen = 0;
u32 len = 0;

volatile u8 miss = 0;

// Double buffer for audio data
u32 rseed = 0;
s8 mixBufferA[MIX_BUFFER_SIZE];
s8 mixBufferB[MIX_BUFFER_SIZE];

volatile bool readA = true;
volatile bool newFrame = true;
u32 mixBufferP;
u16 mixBufferCurr = 0;


/******************************************
 *  FUNCTION DEFINITIONS
 ******************************************/

/******************************************
 *  GAME OBJECT FUNCTIONS
 ******************************************/

void initTargets(void) {
    for( u8 i = 0; i < 4; i++ ) {
        targetidxs[i].idx = -1;
        targetidxs[i].beat = -1;
    }
    for( u8 i = 0; i < KILL_QUEUE_MAX; i++ ) {
        killidxs[i].idx = -1;
        killidxs[i].beat = -1;
        hitidxs[i].idx = -1;
        hitidxs[i].beat = -1;
    }
}

void initObjs(void) {
    for( u8 i = 0; i < 128; i++ ) {
        objs[i].a0 = A0_DISABLE;
        objs[i].a1 = (u16)0;
        objs[i].a2 = (u16)0;
    }
    for( u8 i = 0; i < 26; i++ ) {
        RS_PA(i) = 0x0800 - ((8*i)*0x0700)/MAX_DEPTH;
        RS_PB(i) = (u16)0x0000;
        RS_PC(i) = (u16)0x0000;
        RS_PD(i) = 0x0800 - ((8*i)*0x0700)/MAX_DEPTH;
    }
}

void initEnemies(void) {
    for( int i = 0; i < MAX_ENEMIES; i++ ) {
        enemies[i].next = -2;
        enemies[i].prev = -2;
    }
}

void initMissiles(void) {
    for( int i = 0; i < MAX_MISSILES; i++ ) {
        missiles[i].next = -2;
        missiles[i].prev = -2;
    }
}

u8 findNextBlankEnemy(void) {
    s8 idx = enemybptr;
    if( idx < 0 ) return 0;
    while(enemies[idx].next != -2) {
        idx = (idx + 1) % MAX_ENEMIES;
    }
    return (idx);
}

u8 findNextBlankMissile(void) {
    s8 idx = missilebptr;
    if( idx < 0 ) return 0;
    while(missiles[idx].next != -2) {
        idx = (idx + 1) % MAX_MISSILES;
    }
    return (idx);
}

// Calculate current object positions
void inline calcEnemy(u8 idx) {
    s16 x = enemies[idx].x;
    s16 y = enemies[idx].y;
    s16 z = enemies[idx].z;
    enemies[idx].cx = (((x-120)*z)/MAX_DEPTH + 616) % 512;
    enemies[idx].cy = (((y-80)*z)/MAX_DEPTH + 320) % 256;
}

void inline calcMissile(u8 idx) {
    s16 x = missiles[idx].x;
    s16 y = missiles[idx].y;
    s16 z = missiles[idx].z;
    u8 cz = missiles[idx].cz;
    u8 dz = missiles[idx].dz;
    // Really ugly, should convert to a lookup table or something
    if( cz < dz ) {
        missiles[idx].cx = (((x-120)*(z+(MAX_DEPTH - z)*(cz - z)/(dz - z)))/MAX_DEPTH + 624) % 512;
        missiles[idx].cy = (((y-80)*(z+(MAX_DEPTH - z)*(cz - z)/(dz - z)))/MAX_DEPTH + 328) % 256;
    } else {
        missiles[idx].cx = (((x-120)*(MAX_DEPTH-cz))/(MAX_DEPTH-dz) + 624) % 512;
        missiles[idx].cy = (((y-80)*(MAX_DEPTH-cz))/(MAX_DEPTH-dz) + 328) % 256;
    }
}

void addEnemy(void) {
    if( numEnemies < MAX_ENEMIES ) {
        numEnemies++;
        s8 idx = findNextBlankEnemy();
        if( enemybptr >= 0 ) {
          enemies[enemybptr].next = idx;
        }
        enemies[idx].prev = enemybptr;
        enemies[idx].next = -1;
        enemybptr = idx;
        if( enemyfptr < 0) {
            enemyfptr = idx;
        }
        enemies[idx].z = 20;
        s16 rval = rseed % 9600;
        u8 quad = rval/800;
        enemies[idx].x = QUADX[quad] + (rval % 60);
        enemies[idx].y = QUADY[quad] + ((rval/60) % 40);
        enemies[idx].state = EN_NORMAL;
        calcEnemy(idx);
    }
}

void addMissile(u16 x, u8 y, u8 z) {
    if( numMissiles < MAX_MISSILES ) {
        numMissiles++;
        s8 idx = findNextBlankMissile();
        if( missilebptr >= 0 ) {
          missiles[missilebptr].next = idx;
        }
        missiles[idx].prev = missilebptr;
        missiles[idx].next = -1;
        missilebptr = idx;
        if( missilefptr < 0) {
            missilefptr = idx;
        }
        missiles[idx].x = x;
        missiles[idx].y = y;
        missiles[idx].z = z;
        missiles[idx].dz = (MAX_DEPTH-z)/2 + z;
        missiles[idx].cz = z;
        missiles[idx].state = MS_NORMAL;
        calcMissile(idx);
    }
}

u8 delEnemy(u8 idx) {
    if( enemies[idx].state == EN_TARGET ) hitenemies++;
    totenemies++;
    if(enemyfptr == idx) enemyfptr = enemies[idx].next;
    if(enemybptr == idx) enemybptr = enemies[idx].prev;
    if( enemies[idx].prev >= 0 ) {
        enemies[enemies[idx].prev].next = enemies[idx].next;
    }
    if( enemies[idx].next >= 0 ) {
        enemies[enemies[idx].next].prev = enemies[idx].prev;
    }
    u8 ret = enemies[idx].next;
    enemies[idx].prev = -2;
    enemies[idx].next = -2;
    numEnemies--;
    return ret;
}

u8 delMissile(u8 idx) {
    if( missiles[idx].state == MS_TARGET ) hitmissiles++;
    if(missilefptr == idx) missilefptr = missiles[idx].next;
    if(missilebptr == idx) missilebptr = missiles[idx].prev;
    if( missiles[idx].prev >= 0 ) {
        missiles[missiles[idx].prev].next = missiles[idx].next;
    }
    if( missiles[idx].next >= 0 ) {
        missiles[missiles[idx].next].prev = missiles[idx].prev;
    }
    u8 ret = missiles[idx].next;
    missiles[idx].prev = -2;
    missiles[idx].next = -2;
    numMissiles--;
    
    return ret;
}

void moveEnemies(void) {
    s8 idx = enemyfptr;
    if( numEnemies > 0 ) {
        while( idx >= 0 ) {
            enemies[idx].z += 1;
            if( enemies[idx].z > MAX_DEPTH ) {
                if( enemies[idx].state != EN_NORMAL ) {
                    enemies[idx].z = MAX_DEPTH;
                    idx = enemies[idx].next;
                } else {
                    idx = delEnemy(idx);
                }
            } else {
                calcEnemy(idx);
                if( enemies[idx].z == 60 ) {
                    if( numMissiles < numMaxMissiles ) {
                        addMissile(enemies[idx].x, enemies[idx].y, enemies[idx].z);
                    }
                }
                idx = enemies[idx].next;
            }
        }
    }
}

void moveMissiles(void) {
    s8 idx = missilefptr;
    if( numMissiles > 0 ) {
        while( idx >= 0 ) {
            missiles[idx].cz += 1;
            if( missiles[idx].cz > MAX_DEPTH ) {
                if( missiles[idx].state == MS_NORMAL ) {
                    missiles[idx].state = MS_HIT;
                    tohit++;
                    hitidxs[hitidxback].idx = idx;
                    hitidxback++;
                    hitidxback %= KILL_QUEUE_MAX;
                    tohit++;
                    hitidxs[hitidxback].idx = -1;
                    hitidxback++;
                    hitidxback %= KILL_QUEUE_MAX;
                }
                missiles[idx].cz = MAX_DEPTH;
            } else {
                calcMissile(idx);
            }
            idx = missiles[idx].next;
        }
    }
}

s8 updateEnemy(s8 idx, u8 obj, u8 rs) {
    rs = enemies[idx].z/8;
    OBJ_ENEMY(obj).a0 = (u16)enemies[idx].cy | A0_256PAL | A0_SQUARE | A0_RSENABLE;
    OBJ_ENEMY(obj).a1 = (u16)enemies[idx].cx | A1_SIZE(2) | A1_RS(rs);
    if( enemies[idx].state == EN_TARGET ) OBJ_ENEMY(obj).a2 = S_ENEMY_TARG;
    else OBJ_ENEMY(obj).a2 = S_ENEMY;
    return enemies[idx].next;
}

s8 updateMissile(s8 idx, u8 obj) {
    OBJ_ENEMY(obj).a0 = (u16)missiles[idx].cy | A0_256PAL | A0_SQUARE;
    OBJ_ENEMY(obj).a1 = (u16)missiles[idx].cx | A1_SIZE(1);
    if( missiles[idx].cz < missiles[idx].dz - (MAX_DEPTH - missiles[idx].dz)/3) {
        OBJ_ENEMY(obj).a2 = S_MISSILE_0;
    } else if( missiles[idx].cz < missiles[idx].dz + (MAX_DEPTH - missiles[idx].dz)/3) {
        OBJ_ENEMY(obj).a2 = S_MISSILE_1;
    } else if( missiles[idx].cz % 2) {
        OBJ_ENEMY(obj).a2 = S_MISSILE_2;
    } else {
        OBJ_ENEMY(obj).a2 = S_MISSILE_3;
    }
    if( missiles[idx].state == MS_TARGET ) {
        OBJ_ENEMY(obj).a2 = TARGET_MISSILE(OBJ_ENEMY(obj).a2);
    }
    return missiles[idx].next;
}

void updateObjs(void){
    s8 eidx = enemyfptr;
    s8 midx = missilefptr;
    u8 rs = 0;
    for( u8 i = 0; i < numEnemies+numMissiles; i++ ) {
        if( eidx >= 0 && (numEnemies > 0) ) {
            if( (midx >= 0) && (numMissiles > 0) ) {
                if( enemies[eidx].z > missiles[midx].cz ){
                    eidx = updateEnemy(eidx, i, rs);
                    rs++;
                } else {
                    midx = updateMissile(midx, i);
                }
            } else {
                eidx = updateEnemy(eidx, i, rs);
                rs++;
            }
        } else {
            midx = updateMissile(midx, i);
        }
    }
    for( u8 i = numEnemies+numMissiles; i < OBJ_MAX_ENEMY; i++ ) {
        OBJ_ENEMY(i).a0 = A0_DISABLE;
    }
}

bool targetE(u8 x, u8 y, s8 idx) {
    if( (enemies[idx].state == EN_NORMAL)
      && (enemies[idx].z > 30)) {
        s16 xd = (((enemies[idx].cx + 16) % 512) - (s16)x);
        if( (xd > 0) && (xd < 15) ) {
            s16 yd = (((enemies[idx].cy + 16) % 256) - (s16)y);
            if( (yd > 0) && (yd < 15) ) {
                enemies[idx].state = EN_TARGET;
                targetidxs[targets].idx = idx;
                targetidxs[targets].type = KILL_ENEMY;
                targets++;
                return true;
            }
        }
    }
    return false;
}

bool targetM(u8 x, u8 y, s8 idx) {
    if( (missiles[idx].state == MS_NORMAL)) {
        s16 xd = (((missiles[idx].cx + 8) % 512) - (s16)x);
        if( (xd > 0) && (xd < 15) ) {
            s16 yd = (((missiles[idx].cy + 8) % 256) - (s16)y);
            if( (yd > 0) && (yd < 15) ) {
                missiles[idx].state = MS_TARGET;
                targetidxs[targets].idx = idx;
                targetidxs[targets].type = KILL_MISSILE;
                targets++;
                return true;
            }
        }
    }
    return false;
}

void targetObj(u8 x, u8 y) {
    s8 eidx = enemyfptr;
    s8 midx = missilefptr;
    for( u8 i = 0; i < numEnemies + numMissiles; i++ ) {
        if( eidx >= 0 && (numEnemies > 0) ) {
            if( (midx >= 0) && (numMissiles > 0) ) {
                if( enemies[eidx].z > missiles[midx].cz ){
                    if( targetE(x, y, eidx) ) break;
                    eidx = enemies[eidx].next;
                } else {
                    if( targetM(x, y, midx) ) break;
                    midx = missiles[midx].next;
                }
            } else {
                if( targetE(x, y, eidx) ) break;
                eidx = enemies[eidx].next;
            }
        } else {
            if( targetM(x, y, midx) ) break;
            midx = missiles[midx].next;
        }
    }
}

void killTargets(void) {
    for( u8 i = 0; i < targets; i++ ) {
        killidxs[killidxback] = targetidxs[i];
        killidxback++;
        killidxback %= KILL_QUEUE_MAX;
        if(targetidxs[i].type == KILL_ENEMY) {
            score += 25*targets;
        } else {
            score += 5*targets;
        }
    }
    tokill += targets;
    score += 25*targets*targets;
    targets = 0;
    attacks++;
}

/******************************************
 *  BACKGROUND PALETTE
 ******************************************/

void setPalette(void){
    u16 *pal = pmem;
    pal++;
    u8 r = 8;
    u8 b = 0;
    u8 g = 0;
    for(int i = 0; i < 50; i++) {
        (*pal) = RGB(r, g, b);
        pal++;
    }
    for(r++; r <= 30; r++ ) {
        (*pal) = RGB(r, g, b);
        pal++;
        (*pal) = RGB(r, g, b);
        pal++;
    }
    r = 30;
    for(b+=3; b <= 30; b+=3 ) {
        (*pal) = RGB(r, g, b);
        pal++;
    }
    b = 30;
    for(r-=3; r > 0; r-=3 ) {
        (*pal) = RGB(r, g, b);
        pal++;
    }
    r = 0;
    (*pal) = RGB(r, g, b);
    pal++;
    for(b--; b >= 8; b-- ) {
        (*pal) = RGB(r, g, b);
        pal++;
        (*pal) = RGB(r, g, b);
        pal++;
    }
    b = 8;
    for(int i = 0; i < 50; i++) {
        (*pal) = RGB(r, g, b);
        pal++;
    }

}

/******************************************
 *  BLITTING FUNCTIONS
 ******************************************/

void DrawBitmapDMA( u8 *bmp ) {
  REG_DMA3SAD = (u32)bmp;
  REG_DMA3DAD = (u32)videoBuffer;
  REG_DMA3CNT_L = (u16)(80*240);
  REG_DMA3CNT_H = DMA_ENABLE | DMA_TIMING_IMMEDIATE | DMA_16;
}

void BlitzPal128DMA( u16 *pal ) {
  REG_DMA3SAD = (u32)pal;
  REG_DMA3DAD = (u32)palette;
  REG_DMA3CNT_L = (u16)(128);
  REG_DMA3CNT_H = DMA_ENABLE | DMA_TIMING_IMMEDIATE | DMA_16;
}

void BlitzPalDMA( u16 *pal ) {
  REG_DMA3SAD = (u32)pal;
  REG_DMA3DAD = (u32)palette;
  REG_DMA3CNT_L = (u16)(256);
  REG_DMA3CNT_H = DMA_ENABLE | DMA_TIMING_IMMEDIATE | DMA_16;
}

void BlitzObjTilesDMA( u8 *tiles ) {
  REG_DMA3SAD = (u32)tiles;
  REG_DMA3DAD = (u32)ObjTiles;
  REG_DMA3CNT_L = (u16)(8192);
  REG_DMA3CNT_H = DMA_ENABLE | DMA_TIMING_IMMEDIATE | DMA_16;
}

void BlitzObjPalDMA( u16 *pal ) {
  REG_DMA3SAD = (u32)pal;
  REG_DMA3DAD = (u32)objPalette;
  REG_DMA3CNT_L = (u16)(256);
  REG_DMA3CNT_H = DMA_ENABLE | DMA_TIMING_IMMEDIATE | DMA_16;
}

void writeObjs( void ) {
  REG_DMA3SAD = (u32)objs;
  REG_DMA3DAD = (u32)OAM;
  REG_DMA3CNT_L = (u16)(512);
  REG_DMA3CNT_H = DMA_ENABLE | DMA_TIMING_IMMEDIATE | DMA_16;
}

/******************************************
 *  PAGE FLIPPING
 ******************************************/

void FlipPage(void) {
    if(REG_DISPCNT & VID_BACKBUFFER) {
        videoBuffer = BackBuffer;
    } else {
        videoBuffer = FrontBuffer;
    }
    REG_DISPCNT ^= VID_BACKBUFFER;
}

/******************************************
 *  MIXER FUNCTIONS
 ******************************************/

void initMixer(void) {
    pval = pvala;
    Mixer.beat = 0;
    Mixer.bar = 0;
    Mixer.c = 0;
    Mixer.sfxc = 0;
    Mixer.sfxidx = 0;
    Mixer.cSfx = 0;
    Mixer.cHit = 0;
    Mixer.count = 0;

    for( u8 i = 0; i < Mixer.numSamps; i++ ) {
        Mixer.smps[i].p = Mixer.smps[i].d;
        Mixer.smps[i].c = 0;
    }
    for( u8 i = 0; i < (Mixer.numSfx + Mixer.numHit); i++ ) {
        Mixer.sfx[i].p = Mixer.sfx[i].d;
    }
    for( u8 i = 0; i < 16; i++ ) {
        pvala[i] = 20;
        if( Mixer.mix[0].num > 0 ) pvala[i] += Mixer.smps[Mixer.mix[0].smp1].pval[i];
        if( Mixer.mix[0].num > 1 ) pvala[i] += Mixer.smps[Mixer.mix[0].smp2].pval[i];
        if( Mixer.mix[0].num > 2 ) pvala[i] += Mixer.smps[Mixer.mix[0].smp3].pval[i];
    }
    nextBar = true;
}

// Mix 1 background channel, and SFX
u32 mix1C( u16 len, s8* dst, u8 smp, bool sfx ) {
    s8* p1 = Mixer.smps[smp].p;
    if( sfx ) {
        s8* p2 = Mixer.sfx[Mixer.sfxidx].p;
        for( u16 i = 0; i < len; i++ ) {
            s16 d = *p1 + *p2;
            if( d > 127 ) *dst = 127;
            else if( d < -128 ) (*dst) = -128;
            else *dst = (d & 0x00ff);
            dst++;
            p1++;
            p2++;
        }
        Mixer.sfx[Mixer.sfxidx].p = p2;
    } else {
        for( u16 i = 0; i < len; i++ ) {
            s16 d = *p1;
            if( d > 127 ) *dst = 127;
            else if( d < -128 ) (*dst) = -128;
            else *dst = (d & 0x00ff);
            dst++;
            p1++;
        }
    }
    Mixer.smps[smp].p = p1;
    return (u32)dst;
}

// Mix 2 background channels, and SFX
u32 mix2C( u16 len, s8* dst, u8 smp1, u8 smp2, bool sfx ) {
    s8* p1 = Mixer.smps[smp1].p;
    s8* p2 = Mixer.smps[smp2].p;
    if( sfx ) {
        s8* p3 = Mixer.sfx[Mixer.sfxidx].p;
        for( u16 i = 0; i < len; i++ ) {
            s16 d = *p1 + *p2 + *p3;
            if( d > 127 ) *dst = 127;
            else if( d < -128 ) (*dst) = -128;
            else *dst = (d & 0x00ff);
            dst++;
            p1++;
            p2++;
            p3++;
        }
        Mixer.sfx[Mixer.sfxidx].p = p3;
    } else {
        for( u16 i = 0; i < len; i++ ) {
            s16 d = *p1 + *p2;
            if( d > 127 ) *dst = 127;
            else if( d < -128 ) (*dst) = -128;
            else *dst = (d & 0x00ff);
            dst++;
            p1++;
            p2++;
        }
    }
    Mixer.smps[smp1].p = p1;
    Mixer.smps[smp2].p = p2;
    return (u32)dst;
}

// Mix 3 background channels, and SFX
u32 mix3C( u16 len, s8* dst, u8 smp1, u8 smp2, u8 smp3, bool sfx ) {
    s8* p1 = Mixer.smps[smp1].p;
    s8* p2 = Mixer.smps[smp2].p;
    s8* p3 = Mixer.smps[smp3].p;
    if( sfx ) {
        s8* p4 = Mixer.sfx[Mixer.sfxidx].p;
        for( u16 i = 0; i < len; i++ ) {
            s16 d = *p1 + *p2 + *p3 + *p4;
            if( d > 127 ) *dst = 127;
            else if( d < -128 ) (*dst) = -128;
            else *dst = (d & 0x00ff);
            dst++;
            p1++;
            p2++;
            p3++;
            p4++;
        }
        Mixer.sfx[Mixer.sfxidx].p = p4;
    } else {
        for( u16 i = 0; i < len; i++ ) {
            s16 d = *p1 + *p2 + *p3;
            if( d > 127 ) *dst = 127;
            else if( d < -128 ) (*dst) = -128;
            else *dst = (d & 0x00ff);
            dst++;
            p1++;
            p2++;
            p3++;
        }
    }
    Mixer.smps[smp1].p = p1;
    Mixer.smps[smp2].p = p2;
    Mixer.smps[smp3].p = p3;
    return (u32)dst;
}

u32 mixC( u16 len, u32 dst, u8 cnt, u8 f ) {
    u8 smp1 = Mixer.mix[Mixer.count].smp1;
    u8 smp2 = Mixer.mix[Mixer.count].smp2;
    u8 smp3 = Mixer.mix[Mixer.count].smp3;
    if( cnt == 1 ) {
        dst = mix1C( len, (s8*)dst, smp1, (f == MIX_SFX) );
    } else if( cnt == 2 ) {
        dst = mix2C( len, (s8*)dst, smp1, smp2, (f == MIX_SFX) );
    } else {
        dst = mix3C( len, (s8*)dst, smp1, smp2, smp3, (f == MIX_SFX) );
    }
    if( f == MIX_END) {
        Mixer.smps[smp1].p = Mixer.smps[smp1].d;
        Mixer.smps[smp2].p = Mixer.smps[smp2].d;
        Mixer.smps[smp3].p = Mixer.smps[smp3].d;
    }
    return dst;
}


void nextMixer(void) {
    if( mixBufferP == (u32)mixBufferA ) mixBufferP = (u32)mixBufferB;
    else mixBufferP = (u32)mixBufferA;
    u32 mixB = mixBufferP;
    u16 len = 0;
    // If Sound Effects are needed
    if( (tokill > 0) || (tohit > 0) ) {
        u16 clen = 0;
        while( ((tohit > 0) || (tokill > 0)) && (clen < MIX_BUFFER_SIZE) ) {
            len = MIX_BUFFER_SIZE - clen;
            // The last beat of the sample
            if( (Mixer.c - Mixer.sfxc) > (63*(u32)Mixer.sampleBeat16) ) {
                // The only time sfxc > 0 is when clen == 0
                if( Mixer.sfxc > 0 ) {
                    len = Mixer.sampleBeat16 - Mixer.sfxc;
                    clen += len;
                    mixB = mixC( len, mixB, Mixer.mix[Mixer.count].num, MIX_SFX );
                    Mixer.c += len;
                    Mixer.sfx[Mixer.sfxidx].p = Mixer.sfx[Mixer.sfxidx].d;
                    Mixer.sfxc = 0;
                    if(Mixer.sfxidx < MIX_SFX_KILL_NUM ) {
                        tokill--;
                        kidxmix++;
                        kidxmix %= KILL_QUEUE_MAX;
                        Mixer.cSfx = (Mixer.cSfx + 1) % Mixer.numSfx;
                    } else {
                        tohit--;
                        hidxmix++;
                        hidxmix %= KILL_QUEUE_MAX;
                        Mixer.cHit = (Mixer.cHit + 1) % Mixer.numHit;
                    }
                }
                len = Mixer.sampleSize - Mixer.c;
                clen += len;
                mixB = mixC( len, mixB, Mixer.mix[Mixer.count].num, MIX_END);
                Mixer.c = 0;
                Mixer.count++;
                Mixer.count %= Mixer.loopsize;
            // Currently not on a 16th note mark
            // This will only happen if clen == 0
            } else if((Mixer.c % Mixer.sampleBeat16) > 0){
                // Sound Effect is playing
                if( Mixer.sfxc > 0 ) {
                    len = Mixer.sampleBeat16 - Mixer.sfxc;
                    clen += len;
                    mixB = mixC( len, mixB, Mixer.mix[Mixer.count].num, MIX_SFX );
                    Mixer.sfx[Mixer.sfxidx].p = Mixer.sfx[Mixer.sfxidx].d;
                    Mixer.sfxc = 0;
                    Mixer.c += len;
                    if(Mixer.sfxidx < MIX_SFX_KILL_NUM ) {
                        tokill--;
                        kidxmix++;
                        kidxmix %= KILL_QUEUE_MAX;
                        Mixer.cSfx = (Mixer.cSfx + 1) % Mixer.numSfx;
                    } else {
                        tohit--;
                        hidxmix++;
                        hidxmix %= KILL_QUEUE_MAX;
                        Mixer.cHit = (Mixer.cHit + 1) % Mixer.numHit;
                    }
                // No Sound effect playing
                } else {
                    len = Mixer.sampleBeat16 - (Mixer.c % Mixer.sampleBeat16);
                    clen += len;
                    mixB = mixC( len, mixB, Mixer.mix[Mixer.count].num, MIX_NORMAL);
                    Mixer.c += len;
                }
            // Room for an entire effect
            } else if(len >= Mixer.sampleBeat16) {
                if( tohit > 0 ) {
                    hitidxs[hidxmix].beat = (Mixer.c/Mixer.sampleBeat16) % 16;
                    Mixer.sfxidx = MIX_SFX_HIT(Mixer.cHit);
                } else {
                    killidxs[kidxmix].beat = (Mixer.c/Mixer.sampleBeat16) % 16;
                    Mixer.sfxidx = MIX_SFX_KILL(Mixer.cSfx);
                }
                mixB = mixC( Mixer.sampleBeat16, mixB, Mixer.mix[Mixer.count].num, MIX_SFX);
                Mixer.sfx[Mixer.sfxidx].p = Mixer.sfx[Mixer.sfxidx].d;
                Mixer.sfxc = 0;
                Mixer.c += Mixer.sampleBeat16;
                clen += Mixer.sampleBeat16;
                if(Mixer.sfxidx < MIX_SFX_KILL_NUM ) {
                    tokill--;
                    kidxmix++;
                    kidxmix %= KILL_QUEUE_MAX;
                    Mixer.cSfx = (Mixer.cSfx + 1) % Mixer.numSfx;
                } else {
                    tohit--;
                    hidxmix++;
                    hidxmix %= KILL_QUEUE_MAX;
                    Mixer.cHit = (Mixer.cHit + 1) % Mixer.numHit;
                }
            // Not enough room for an entire effect, mix in partial
            } else {
                if( tohit > 0 ) {
                    hitidxs[hidxmix].beat = (Mixer.c/Mixer.sampleBeat16) % 16;
                    Mixer.sfxidx = MIX_SFX_HIT(Mixer.cHit);
                } else {
                    killidxs[kidxmix].beat = (Mixer.c/Mixer.sampleBeat16) % 16;
                    Mixer.sfxidx = MIX_SFX_KILL(Mixer.cSfx);
                }
                mixB = mixC( len, mixB, Mixer.mix[Mixer.count].num, MIX_SFX);
                Mixer.c += len;
                clen += len;
                Mixer.sfxc += len;
            }
        }
        // Mix the remaining audio
        if( clen < MIX_BUFFER_SIZE ) {
            len = MIX_BUFFER_SIZE - clen;
            if( Mixer.c + len >= Mixer.sampleSize ) {
                len = Mixer.sampleSize - Mixer.c;
                mixB = mixC( len, mixB, Mixer.mix[Mixer.count].num, MIX_END);
                Mixer.c = 0;
                Mixer.count++;
                Mixer.count %= Mixer.loopsize;
                len = MIX_BUFFER_SIZE - clen - len;
                mixC( len, mixB, Mixer.mix[Mixer.count].num, MIX_NORMAL);
                Mixer.c += len;
            } else {
                mixC( len, mixB, Mixer.mix[Mixer.count].num, MIX_NORMAL);
                Mixer.c += len;
            }
        }
    } else if( Mixer.c + MIX_BUFFER_SIZE >= Mixer.sampleSize ) {
        len = Mixer.sampleSize - Mixer.c;
        mixB = mixC( len, mixB, Mixer.mix[Mixer.count].num, MIX_END);
        Mixer.c = 0;
        Mixer.count++;
        Mixer.count %= Mixer.loopsize;
        len = MIX_BUFFER_SIZE - len;
        mixC( len, mixB, Mixer.mix[Mixer.count].num, MIX_NORMAL);
        Mixer.c += len;
    } else {
        mixC( MIX_BUFFER_SIZE, mixB, Mixer.mix[Mixer.count].num, MIX_NORMAL);
        Mixer.c += MIX_BUFFER_SIZE;
    }
}

// Calculate the palette shift values for the next sample
void nextPval(void) {
    u8 offset = (Mixer.bar+1)*16 % 64;
    u8 c = Mixer.count;
    if( offset == 0 ) {
        c = c + 1 % Mixer.loopsize;
    }
    if( pval == pvala ) {
        for( u8 i = 0; i < 16; i++ ) {
            pvalb[i] = 20;
            if( Mixer.mix[c].num > 0 ) pvalb[i] +=
                    Mixer.smps[Mixer.mix[c].smp1].pval[offset + i];
            if( Mixer.mix[c].num > 1 ) pvalb[i] +=
                    Mixer.smps[Mixer.mix[c].smp2].pval[offset + i];
            if( Mixer.mix[c].num > 2 ) pvalb[i] +=
                    Mixer.smps[Mixer.mix[c].smp3].pval[offset + i];
        }
    } else {
        for( u8 i = 0; i < 16; i++ ) {
            pvala[i] = 20;
            if( Mixer.mix[c].num > 0 ) pvala[i] +=
                    Mixer.smps[Mixer.mix[c].smp1].pval[offset + i];
            if( Mixer.mix[c].num > 1 ) pvala[i] +=
                    Mixer.smps[Mixer.mix[c].smp2].pval[offset + i];
            if( Mixer.mix[c].num > 2 ) pvala[i] +=
                    Mixer.smps[Mixer.mix[c].smp3].pval[offset + i];
        }
    }
}

// Wait for the next Screen Refresh
void waitVBlank(void) {
    while( *vcount >= 160 );
    while( *vcount < 160 );
}

/******************************************
 *  GAME FUNCTION
 ******************************************/

void playgame(void) {
    bool nextBG = false;
    bool oldBG = false;
    //u8 health = 5;
    u8 targetx = 112;
    u8 targety = 72;
    u8 selfx = 104;
    u8 selfy = 72;
    u16 linez = 0;
    u8 movedelay = 0;
    u8 movefactor = 1;
    s8 oldbeat = -1;
    score = 0;
    pcnt = 20;
    health = 5;
    targets = 0;
    tokill = 0;
    killidx = 0;
    kidxmix = 0;
    killidxback = 0;
    tohit = 0;
    hitidx = 0;
    hidxmix = 0;
    hitidxback = 0;
    enemyfptr = -1;
    enemybptr = -1;
    numEnemies = 0;
    missilefptr = -1;
    missilebptr = -1;
    numMissiles = 0;
    hitmissiles = 0;
    hitenemies = 0;
    totenemies = 0;
    attacks = 0;
    u8 whirlnum = 0;
    u8 whirlcnt = 0;

    btns = *BUTTONS;

    initObjs();
    initEnemies();
    initMissiles();
    initTargets();
    initMixer();
    // create background palettes
    setPalette();

    // setup Sound Double Buffer
    mixBufferP = (u32)mixBufferA;
    SetMode(4 | VID_BG2 | VID_OBJ);
    //SetMode(4 | VID_BG2);
    REG_BLDCNT = BLD_2_BG2 | BLD_1_OBJ;
    REG_BLDALPHA = 0x000B | 0x1000;

    // initialize the background palette
    BlitzPal128DMA(&pmem[20]);
    DrawBitmapDMA(whirl[0]);
    FlipPage();
    DrawBitmapDMA(whirl[0]);
    BlitzObjPalDMA(objpal);
    BlitzObjTilesDMA(objtiles);


    // Initilize crosshairs
    OBJ_TARGET.a0 = targety | A0_256PAL | A0_SQUARE | A0_SEMITRANS;
    OBJ_TARGET.a1 = targetx | A1_SIZE(1);
    OBJ_TARGET.a2 = S_TARGET;

    // Initilize floating self
    OBJ_SELF.a0 = selfy | A0_256PAL | A0_SQUARE | A0_SEMITRANS;
    OBJ_SELF.a1 = selfx | A1_SIZE(2);
    OBJ_SELF.a2 = S_SELF;

    // Initilize score objects
    for( u16 i = 0; i < 10; i++ ) {
        OBJ_NUM(i).a0 = (u16)(160-8) | A0_256PAL | A0_SQUARE;
        OBJ_NUM(i).a1 = (u16)(240-(6*(i+1))) | A1_SIZE(0);
        OBJ_NUM(i).a2 = S_NUM(0);
    }
    // Initilize health objects
    for( u16 i = 0; i < 5; i++ ) {
        OBJ_HEALTH(i).a0 = (u16)(160-8) | A0_256PAL | A0_SQUARE;
        //OBJ_HEALTH(i).a0 = (u16)(255) | A0_256PAL | A0_SQUARE;
        OBJ_HEALTH(i).a1 = (u16)(4 + i*8) | A1_SIZE(0);
    }
    OBJ_HEALTH(0).a2 = S_HEALTH_ON0;
    OBJ_HEALTH(1).a2 = S_HEALTH_ON1;
    OBJ_HEALTH(2).a2 = S_HEALTH_ON2;
    OBJ_HEALTH(3).a2 = S_HEALTH_ON3;
    OBJ_HEALTH(4).a2 = S_HEALTH_ON4;

    writeObjs();

    // Initilize Interrupt Handler
    REG_IME = 0x00;
    REG_INTERRUPT = (u32)IntHandler;
    REG_IE |= INT_TIM1 | INT_TIM2 | INT_VBLANK;
    REG_DISPSTAT |= 0x08;
    REG_IME = 0x01;

    // Initialize sound DAC
    REG_DMA1CNT_H = 0;
    REG_SGCNT0_H = DSOUND_A_RIGHT_CHANNEL | DSOUND_A_LEFT_CHANNEL | DSOUND_A_FIFO_RESET;
    REG_SGCNT1 = SOUND_MASTER_ENABLE;

    // Seed the audio buffer
    nextMixer();

    // Transfer the sound buffer to the Sound Queue
    REG_DMA1SAD = (u32)mixBufferP;
    REG_DMA1DAD = SOUND_FIFO_A;

    REG_DMA1CNT_H = DMA_DEST_FIXED | DMA_REPEAT | DMA_32 |
        DMA_SPECIAL | DMA_ENABLE;
    readA = true;

    // Initilize the Timers
    REG_TM0D = 0xffff;
    REG_TM1D = 0xffff - MIX_BUFFER_SIZE;
    REG_TM2D = 0xffff - Mixer.sampleBeat;

    // Start the Timers
    REG_TM1CNT = TIMER_ENABLE | TIMER_SCALE_1024 | TIMER_OVERFLOW;
    REG_TM0CNT = TIMER_ENABLE | TIMER_SCALE_1024;
    REG_TM2CNT = TIMER_ENABLE | TIMER_SCALE_1024 | TIMER_OVERFLOW;

    // Main Game Loop
    while(true)
    {
        // No random number generator in the GBA so must improvise
        rseed+= 1897;
        waitVBlank();

        // Calculate the current music beat
        // Shift Background Palette based on music beat
        for( u16 i = 0; i < 16; i++ ) {
            if( *beatTimer < 0xffff-Mixer.sampleBeat+(((i+1)*Mixer.sampleBeat)/16) ) {
                if(Mixer.beat != i) {
                    pcnt = pval[Mixer.beat];
                    Mixer.beat = i;
                }
                break;
            }
        }
        if(Mixer.beat != oldbeat) {
            oldbeat = Mixer.beat;
            OBJ_SELF.a2 = S_SELF;
        }

        // Blitz the shifted palette to memory
        BlitzPal128DMA(&pmem[pcnt]);
        writeObjs();

        // Animate the background
        whirlcnt += miss;
        if( whirlcnt >= 5 ) {
            whirlcnt = 0;
            whirlnum--;
            whirlnum %= 16;
            nextBG = true;
        }

        FlipPage();

        // BUTTON SEGMENT
            u16 change = (*BUTTONS) ^ btns;
            if( change & BTN_A ) {
                if( btns & BTN_A ) {
                    Mixer.smps[0].v = (Mixer.smps[0].v + 1) % 3;
                }
            }
            if( change & BTN_B ) {
                if( btns & BTN_B ) {
                    Mixer.smps[1].v = (Mixer.smps[1].v + 1) % 3;
                }
            }
            if( change & BTN_R ) {
                if( btns & BTN_R ) {
                    OBJ_TARGET.a2 = S_TARGETING;
                } else {
                    OBJ_TARGET.a2 = S_TARGET;
                    killTargets();
                }
            }
            if( !((*BUTTONS) & BTN_R) ) {
                if( targets < 4 ) {
                    targetObj(targetx, targety);
                }
            }
            if( change & BTN_L ) {
                if( btns & BTN_L ) {
                    movefactor = 2;
                } else {
                    movefactor = 1;
                }
            }
            if( !((*BUTTONS) & BTN_UP) ) {
                if( targety > 0 ) {
                    targety-=2*miss*movefactor;
                    if( targety > 160 ) targety = 0;
                }
                OBJ_TARGET.a0 = targety | A0_256PAL | A0_SQUARE | A0_SEMITRANS;
            }
            if( !((*BUTTONS) & BTN_DOWN) ) {
                if( targety < 160-15 ) {
                    targety+=2*miss*movefactor;
                    if( targety > 160-15 ) targety = 160-15;
                }
                OBJ_TARGET.a0 = targety | A0_256PAL | A0_SQUARE | A0_SEMITRANS;
            }
            if( !((*BUTTONS) & BTN_LEFT) ) {
                rseed+=1273;
                if( targetx > 0 ) {
                    targetx-=2*miss*movefactor;
                    if( targetx > 240 ) targetx = 0;
                }
                OBJ_TARGET.a1 = targetx | A1_SIZE(1);
            }
            if( !((*BUTTONS) & BTN_RIGHT) ) {
                rseed+=8723;
                if( targetx < 240-15 ) {
                    targetx+=2*miss*movefactor;
                    if( targetx > 240-15 ) targetx = 240-15;
                }
                OBJ_TARGET.a1 = targetx | A1_SIZE(1);
            }
            if( change & BTN_SELECT ) {
                if( btns & BTN_SELECT ) {
                    if( health < 5) health++;
                }
            }
            if( change & BTN_START ) {
                if( btns & BTN_START ) {
                    break;
                }
            }
            btns ^= change;

        if(readA){
            // Mix next audio buffer
            nextMixer();
            readA = false;
            
            // Make the hero float around
            if( rseed % 5 < 2 ) { if(selfx < 106) selfx++; }
            else if( rseed % 5 > 2) { if(selfx > 102) selfx--; }
            if( rseed % 3 < 1 ) { if(selfy < 73) selfy++; }
            else if( rseed % 3 > 1) { if(selfy > 71) selfy--; }

            OBJ_SELF.a0 = selfy | A0_256PAL | A0_SQUARE | A0_SEMITRANS;
            OBJ_SELF.a1 = selfx |  A1_SIZE(2);

        }
        // Generate next set of palette shifts
        if(nextBar){
            nextPval();
            nextBar = false;
        }

        // Draw the background image
        if( oldBG ) {
            DrawBitmapDMA(whirl[whirlnum]);
            oldBG = false;
        }
        if( nextBG ) {
            DrawBitmapDMA(whirl[whirlnum]);
            nextBG = false;
            oldBG = true;
        }
        movedelay += miss;
        if( movedelay > 3 ) {
            linez = (linez + 1) % MAX_DEPTH;
        }
            // Draw the lines on the background image
            for( u16 i = 0; i < 4; i++) {
                u16 s = 68*((57*i + linez) % MAX_DEPTH)/MAX_DEPTH + 8;
                u16 f = 68*((57*i + 127 + linez) % MAX_DEPTH)/MAX_DEPTH + 8;
                if( s < f ) {
                    for( u8 j = 8; j < s; j++ ) {
                        videoBuffer[linea[i][j]] = ((u16*)whirl[whirlnum])[linea[i][j]];
                    }
                    for( u8 j = s; j < f; j++ ) {
                        if( ((j >> 1) + i + 1) % 2 ) {
                           videoBuffer[linea[i][j]] = (u16)(whirl[whirlnum][linea[i][j]*2] + (linecol(j) << 8));
                        } else {
                            videoBuffer[linea[i][j]] = (u16)(linecol(j) + (whirl[whirlnum][linea[i][j]*2] << 8));
                        }
                    }
                    for( u8 j = f; j < 76; j++ ) {
                        videoBuffer[linea[i][j]] = ((u16*)whirl[whirlnum])[linea[i][j]];
                    }
                } else {
                    for( u8 j = 8; j < f; j++ ) {
                        if( ((j >> 1) + i + 1) % 2 ) {
                           videoBuffer[linea[i][j]] = (u16)(whirl[whirlnum][linea[i][j]*2] + (linecol(j) << 8));
                        } else {
                            videoBuffer[linea[i][j]] = (u16)(linecol(j) + (whirl[whirlnum][linea[i][j]*2] << 8));
                        }
                    }
                    for( u8 j = f; j < s; j++ ) {
                        videoBuffer[linea[i][j]] = ((u16*)whirl[whirlnum])[linea[i][j]];
                    }
                    for( u8 j = s; j < 76; j++ ) {
                        if( ((j >> 1) + i + 1) % 2 ) {
                           videoBuffer[linea[i][j]] = (u16)(whirl[whirlnum][linea[i][j]*2] + (linecol(j) << 8));
                        } else {
                            videoBuffer[linea[i][j]] = (u16)(linecol(j) + (whirl[whirlnum][linea[i][j]*2] << 8));
                        }
                    }
                }
            }
            for( u8 i = 0; i < 4; i++) {
                u16 s = 50*((23*i + linez) % MAX_DEPTH)/MAX_DEPTH + 8;
                u16 f = 50*((23*i + 89 + linez) % MAX_DEPTH)/MAX_DEPTH + 8;
                if( s < f ) {
                    for( u8 j = 4; j < s; j++ ) {
                        videoBuffer[lineb[i][j]] = ((u16*)whirl[whirlnum])[lineb[i][j]];
                    }
                    for( u8 j = s; j < f; j++ ) {
                        videoBuffer[lineb[i][j]] = (u16)(linecol(j) + (linecol(j) << 8));
                    }
                    for( u8 j = f; j < 58; j++ ) {
                        videoBuffer[lineb[i][j]] = ((u16*)whirl[whirlnum])[lineb[i][j]];
                    }
                } else {
                    for( u8 j = 4; j < f; j++ ) {
                        videoBuffer[lineb[i][j]] = (u16)(linecol(j) + (linecol(j) << 8));
                    }
                    for( u8 j = f; j < s; j++ ) {
                        videoBuffer[lineb[i][j]] = ((u16*)whirl[whirlnum])[lineb[i][j]];
                    }
                    for( u8 j = s; j < 58; j++ ) {
                        videoBuffer[lineb[i][j]] = (u16)(linecol(j) + (linecol(j) << 8));
                    }
                }
            }

        // Draw Score
        u32 stmp = score;
        for( u8 i = 0; i++; i < 10 ) {
            OBJ_NUM(i).a2 = S_NUM(stmp % 10);
            stmp /= 10;
        }

        //Draw Health Bar
        if(health > 0) OBJ_HEALTH(0).a2 = S_HEALTH_ON0;
        else OBJ_HEALTH(0).a2 = S_HEALTH_OFF0;
        if(health > 1) OBJ_HEALTH(1).a2 = S_HEALTH_ON1;
        else OBJ_HEALTH(1).a2 = S_HEALTH_OFF1;
        if(health > 2) OBJ_HEALTH(2).a2 = S_HEALTH_ON2;
        else OBJ_HEALTH(2).a2 = S_HEALTH_OFF2;
        if(health > 3) OBJ_HEALTH(3).a2 = S_HEALTH_ON3;
        else OBJ_HEALTH(3).a2 = S_HEALTH_OFF3;
        if(health > 4) OBJ_HEALTH(4).a2 = S_HEALTH_ON4;
        else OBJ_HEALTH(4).a2 = S_HEALTH_OFF4;

        numMaxEnemies = maxE[Mixer.count];
        numMaxMissiles = maxM[Mixer.count];

        // Check if an enemy is being shot
        if(killidxs[killidx].beat == Mixer.beat ) {
            if(killidxs[killidx].type == KILL_ENEMY)
                delEnemy(killidxs[killidx].idx);
            else
                delMissile(killidxs[killidx].idx);
            killidxs[killidx].beat = -1;
            killidx++;
            killidx %= KILL_QUEUE_MAX;
            OBJ_SELF.a2 = S_SELF_KILL(Mixer.beat % 2);
        }
        // Check if you're being hit
        if(hitidxs[hitidx].beat == Mixer.beat ) {
            if(hitidxs[hitidx].idx >= 0 ) {
                delMissile(hitidxs[hitidx].idx);
                health--;
            } 
            hitidxs[hitidx].beat = -1;
            hitidx++;
            hitidx %= KILL_QUEUE_MAX;
            OBJ_SELF.a2 = S_SELF_HIT;
        }
        // Move the enemies/missiles
        if( movedelay > 3 ) {
            movedelay = 0;
            moveEnemies();
            moveMissiles();
        }
        updateObjs();
        if((numEnemies < numMaxEnemies) && ((rseed % 23) == 0)) addEnemy();

        miss = 0;
        if(health == 0) break;
    } // GAME LOOP


    // Turn off timers
    REG_SGCNT1 = 0;
    REG_TM1CNT = 0;
    REG_TM0CNT = 0;
    REG_TM2CNT = 0;

    // Perform fade to black
    REG_BLDCNT = BLD_1_BG2 | BLD_1_OBJ | BLD_DARK;
    for( u8 i = 0; i <= 16; i++ ) {
        REG_BLDY = i;
        miss = 0;
        while( miss < 5 ) { }
    }

    // Draw Game Over Screen and Stats
    btns = *BUTTONS;

    initObjs();
    BlitzObjPalDMA(menupal);
    BlitzObjTilesDMA(menuitems);

    SetMode(4 | VID_BG2 | VID_OBJ);
    REG_BLDCNT = BLD_2_BG2 | BLD_1_OBJ;
    FlipPage();
    DrawBitmapDMA(gameover);
    BlitzPalDMA(gameoverpal);
    FlipPage();

    // see if a new high score was made, and update the highscore list
    bool newhighscore = false;
    u32 stmp = score;
    for( int i = 0; i < 10; i++ ) {
        if( stmp > highscores[i]) {
            u32 tmpscore = highscores[i];
            highscores[i] = stmp;
            stmp = tmpscore;
            newhighscore = true;
        }
    }
    WriteHighScores();

    // Draw the stats
    stmp = score;
    for( u16 i = 0; i < 10; i++ ) {
        OBJ_NUM(i).a0 = (u16)29 | A0_256PAL | A0_SQUARE;
        OBJ_NUM(i).a1 = (u16)(126+(6*(9-i))) | A1_SIZE(0);
        OBJ_NUM(i).a2 = S_NUM(stmp % 10);
        stmp /= 10;
    }
    u16 mtmp = hitmissiles;
    for( u16 i = 0; i < 5; i++ ) {
        OBJ(i+10).a0 = (u16)51 | A0_256PAL | A0_SQUARE;
        OBJ(i+10).a1 = (u16)(126+(6*(4-i))) | A1_SIZE(0);
        OBJ(i+10).a2 = S_NUM(mtmp % 10);
        mtmp /= 10;
    }
    u16 etmp = hitenemies;
    for( u16 i = 0; i < 5; i++ ) {
        OBJ(i+15).a0 = (u16)73 | A0_256PAL | A0_SQUARE;
        OBJ(i+15).a1 = (u16)(126+(6*(4-i))) | A1_SIZE(0);
        OBJ(i+15).a2 = S_NUM(etmp % 10);
        etmp /= 10;
    }
    u16 hprc = (hitenemies*100)/totenemies;
    for( u16 i = 0; i < 3; i++ ) {
        OBJ(i+20).a0 = (u16)95 | A0_256PAL | A0_SQUARE;
        OBJ(i+20).a1 = (u16)(126+(6*(2-i))) | A1_SIZE(0);
        OBJ(i+20).a2 = S_NUM(hprc % 10);
        hprc /= 10;
    }
    u16 avg = (hitenemies+hitmissiles)*100/attacks;
    for( u16 i = 0; i < 4; i++ ) {
        OBJ(i+25).a0 = (u16)118 | A0_256PAL | A0_SQUARE;
        OBJ(i+25).a1 = (u16)(126+(6*(3-i))) | A1_SIZE(0);
        if( i == 2 ) {
            OBJ(i+25).a2 = S_NUM(10);
        } else {
            OBJ(i+25).a2 = S_NUM(avg % 10);
            avg /= 10;
        }
    }
    if( newhighscore ) {
        OBJ(31).a0 = 140 | A0_256PAL | A0_HORZ;
        OBJ(31).a1 = 71 | A1_SIZE(2);
        OBJ(31).a2 = 704;
        OBJ(32).a0 = 140 | A0_256PAL | A0_HORZ;
        OBJ(32).a1 = 103 | A1_SIZE(2);
        OBJ(32).a2 = 712;
        OBJ(33).a0 = 140 | A0_256PAL | A0_HORZ;
        OBJ(33).a1 = 135 | A1_SIZE(2);
        OBJ(33).a2 = 720;
    }

    writeObjs();

    while(true) {
        u16 change = (*BUTTONS) ^ btns;
        if( change & BTN_A ) {
            if( btns & BTN_A ) {
                break;
            }
        }
        btns ^= change;
    }
}

void highscoremenu(void){
    u16 btns = *BUTTONS;
    initObjs();

    SetMode(4 | VID_BG2 | VID_OBJ);
    REG_BLDCNT = BLD_2_BG2 | BLD_1_OBJ;
    // Draw Background
    FlipPage();
    DrawBitmapDMA(highscore);
    BlitzPalDMA(highscorepal);
    FlipPage();
    BlitzObjPalDMA(objpal);
    BlitzObjTilesDMA(objtiles);

    // Draw high scores
    for( u16 j = 0; j < 10; j++ ) {
        u32 stmp = highscores[j];
        for( u16 i = 0; i < 10; i++ ) {
            u16 idx = (j+1)*10-1-i;
            OBJ_NUM(idx).a0 = (u16)(20+j*14) | A0_256PAL | A0_SQUARE;
            OBJ_NUM(idx).a1 = (u16)(90+(6*(9-i))) | A1_SIZE(0);
            OBJ_NUM(idx).a2 = S_NUM(stmp % 10);
            stmp /= 10;
        }
    }
    writeObjs();

    while(true) {
        u16 change = (*BUTTONS) ^ btns;
        if( change & BTN_A ) {
            if( btns & BTN_A ) {
                break;
            }
        }
        btns ^= change;
    }
}

// Cycle through help menus
void helpmenu(void){
    u16 btns = *BUTTONS;
    u8 help = 0;
    bool newpage = false;

    SetMode(4 | VID_BG2 );
    REG_BLDCNT = BLD_2_BG2 | BLD_1_OBJ;
    FlipPage();
    DrawBitmapDMA(helpmenubmp[help]);
    BlitzPalDMA(helppal[help]);
    FlipPage();
    
    while(true) {
        u16 change = (*BUTTONS) ^ btns;
        if( change & BTN_A ) {
            if( btns & BTN_A ) {
                newpage = true;
                help = (help + 1) % numhelp;
            }
        }
        if( change & BTN_B ) {
            if( btns & BTN_B ) {
                break;
            }
        }
        btns ^= change;
        if( newpage ) {
            newpage = false;
            DrawBitmapDMA(helpmenubmp[help]);
            BlitzPalDMA(helppal[help]);
            FlipPage();
        }
    }

}

void mainmenu(void){
    u8 menu = 0;

    while(true) {
        u16 btns = *BUTTONS;

        initObjs();
        writeObjs();
        SetMode(4 | VID_BG2 | VID_OBJ);
        REG_BLDCNT = BLD_2_BG2 | BLD_1_OBJ;
        FlipPage();
        DrawBitmapDMA(title);
        BlitzPalDMA(titlepal);
        FlipPage();

        BlitzObjPalDMA(menupal);
        BlitzObjTilesDMA(menuitems);

        while(true) {
            if( menu == 0 ) {
                OBJ(0).a0 = 73 | A0_256PAL | A0_HORZ;
                OBJ(0).a1 = 31 | A1_SIZE(2);
                OBJ(0).a2 = 512;
                OBJ(1).a0 = A0_DISABLE;
                OBJ(2).a0 = A0_DISABLE;
                OBJ(3).a0 = A0_DISABLE;
                OBJ(4).a0 = A0_DISABLE;
            } else if( menu == 1 ) {
                OBJ(1).a0 = 92 | A0_256PAL | A0_HORZ;
                OBJ(1).a1 = 31 | A1_SIZE(2);
                OBJ(1).a2 = 576;
                OBJ(2).a0 = 92 | A0_256PAL | A0_HORZ;
                OBJ(2).a1 = 63 | A1_SIZE(2);
                OBJ(2).a2 = 584;
                OBJ(3).a0 = 92 | A0_256PAL | A0_HORZ;
                OBJ(3).a1 = 95 | A1_SIZE(2);
                OBJ(3).a2 = 592;
                OBJ(0).a0 = A0_DISABLE;
                OBJ(4).a0 = A0_DISABLE;
            } else {
                OBJ(4).a0 = 110 | A0_256PAL | A0_HORZ;
                OBJ(4).a1 = 31 | A1_SIZE(2);
                OBJ(4).a2 = 640;
                OBJ(0).a0 = A0_DISABLE;
                OBJ(1).a0 = A0_DISABLE;
                OBJ(2).a0 = A0_DISABLE;
                OBJ(3).a0 = A0_DISABLE;
            }
            waitVBlank();
            writeObjs();
            u16 change = (*BUTTONS) ^ btns;
            if( change & BTN_DOWN ) {
                if( btns & BTN_DOWN ) {
                    menu = (menu + 1) % 3;
                }
            }
            if( change & BTN_UP ) {
                if( btns & BTN_UP ) {
                    menu = (menu + 2) % 3;
                }
            }
            if( change & BTN_A ) {
                if( btns & BTN_A ) {
                    if( menu == 0 ) {
                        playgame();
                    } else if( menu == 1 ) {
                        highscoremenu();
                    } else {
                        helpmenu();
                    }

                    break;
                }
            }
            btns = *BUTTONS;
        }
    }

}

/******************************************
 *  MAIN FUNCTION
 ******************************************/

int main(void)
{
    // Read High Scores from SRAM
    ReadHighScores();
    // Enter the menu system
    mainmenu();

    return 0;
}

/* END OF FILE */
